"""Module for miscellaneous functions and methods"""
import sys

from pyvista.utilities.errors import GPUInfo
import scooby

import numpy as np
import vtk

VTK9 = vtk.vtkVersion().GetVTKMajorVersion() >= 9

def vtk_cell_info(grid):
    """Returns version consistent connectivity and cell offset arrays.

    Notes
    -----
    VTKv9 changed the connectivity and offset arrays:

    Topology:
    ---------
    Cell 0: Triangle | point ids: {0, 1, 2}
    Cell 1: Triangle | point ids: {5, 7, 2}
    Cell 2: Quad     | point ids: {3, 4, 6, 7}
    Cell 4: Line     | point ids: {5, 8}

    VTKv9
    =====
    Offsets:      {0, 3, 6, 10, 12}
    Connectivity: {0, 1, 2, 5, 7, 2, 3, 4, 6, 7, 5, 8}

    Prior to VTKv9
    ==============
    Offsets:      {0, 4, 8, 13, 16}
    Connectivity: {3, 0, 1, 2, 3, 5, 7, 2, 4, 3, 4, 6, 7, 2, 5, 8}

    """
    if VTK9:
        # for pyvista < 0.25.0
        if not hasattr(grid, 'cell_connectivity'):
            carr = grid.GetCells()
            cells = vtk.util.numpy_support.vtk_to_numpy(carr.GetConnectivityArray())
        else:
            cells = grid.cell_connectivity
        offset = grid.offset - 1
    else:
        cells, offset = grid.cells, grid.offset

    if cells.dtype != np.int64:
        cells = cells.astype(np.int64)

    if offset.dtype != np.int64:
        offset = offset.astype(np.int64)

    return cells, offset


def kill_process(proc_pid):
    """Kill a process with extreme prejudice"""
    import psutil  # imported here to avoid import errors when unused
    process = psutil.Process(proc_pid)
    for proc in process.children(recursive=True):
        proc.kill()
    process.kill()


class Report(scooby.Report):
    """A class for custom scooby.Report."""

    def __init__(self, additional=None, ncol=3, text_width=80, sort=False,
                 gpu=True):
        """Generate a :class:`scooby.Report` instance.

        Parameters
        ----------
        additional : list(ModuleType), list(str)
            List of packages or package names to add to output information.

        ncol : int, optional
            Number of package-columns in html table; only has effect if
            ``mode='HTML'`` or ``mode='html'``. Defaults to 3.

        text_width : int, optional
            The text width for non-HTML display modes

        sort : bool, optional
            Alphabetically sort the packages

        gpu : bool
            Gather information about the GPU. Defaults to ``True`` but if
            experiencing renderinng issues, pass ``False`` to safely generate
            a report.

        """
        # Mandatory packages.
        core = ['pyansys', 'pyvista', 'vtk', 'numpy',
                'appdirs', 'psutil', 'pexpect']

        if sys.platform != 'darwin':
            core.append('pyansys')

        # Optional packages.
        optional = ['scipy', 'matplotlib']

        # Information about the GPU - bare except in case there is a rendering
        # bug that the user is trying to report.
        if gpu:
            try:
                extra_meta = [(t[1], t[0]) for t in GPUInfo().get_info()]
            except:
                extra_meta = ("GPU Details", "error")
        else:
            extra_meta = ("GPU Details", "None")

        scooby.Report.__init__(self, additional=additional, core=core,
                               optional=optional, ncol=ncol,
                               text_width=text_width, sort=sort,
                               extra_meta=extra_meta)



"""CORBA implementation of the MAPDL interface"""
import atexit
from threading import Thread
import time
import re
import os
import subprocess

from pyansys.mapdl import _Mapdl
from pyansys.misc import kill_process

try:
    from ansys_corba import CORBA
except:
    raise ImportError('Missing ansys_corba libaries.\n' +
                      'Only supported on Python3.5 - Python3.8 for '
                      'Linux and Windows\n')

INSTANCES = []

# Windows has issues when closing
@atexit.register
def cleanup():
    if os.name == 'nt':
        for instance in INSTANCES:
            instance.kill()


def threaded(fn):
    """ calls a function using a thread """
    def wrapper(*args, **kwargs):
        thread = Thread(target=fn, args=args, kwargs=kwargs)
        thread.start()
        return thread
    return wrapper


def tail(filename, nlines):
    """ Read the last nlines of a text file """
    with open(filename) as qfile:
        qfile.seek(0, os.SEEK_END)
        endf = position = qfile.tell()
        linecnt = 0
        while position >= 0:
            qfile.seek(position)
            next_char = qfile.read(1)
            if next_char == "\n" and position != endf-1:
                linecnt += 1

            if linecnt == nlines:
                break
            position -= 1

        if position < 0:
            qfile.seek(0)

        return qfile.read()


class MapdlCorba(_Mapdl):
    """CORBA implementation of the MAPDL interface"""

    def __init__(self, exec_file, run_location,
                 jobname='file', nproc=2, override=False,
                 loglevel='INFO', additional_switches='',
                 start_timeout=120, interactive_plotting=False,
                 check_version=True,
                 prefer_pexpect=True, log_apdl='w',
                 log_broadcast=False):
        self._broadcast_logger = None
        self._server = None
        self._log_broadcast = log_broadcast

        # CORBA/AAS was introduced in v17
        version = int(re.findall(r'\d\d\d', exec_file)[0])
        if version < 170:
            raise RuntimeError('MAPDL AAS CORBA server requires '
                               'v17.0 or greater.')

        # this will launch MAPDL
        super().__init__(exec_file, run_location, jobname, nproc,
                         override, loglevel, additional_switches,
                         start_timeout, interactive_plotting,
                         log_apdl)
        INSTANCES.append(self)

    @property
    def _broadcast_file(self):
        return os.path.join(self.path, 'mapdl_broadcasts.txt')

    def _launch(self):
        """Open a connection to ANSYS via a CORBA interface"""
        # Using stored parameters so launch command can be run from a
        # cached state (when launching the GUI)
        self._log.info('Connecting to ANSYS via CORBA')

        # create a dummy input file for getting NON-INTERACTIVE without
        # running /BATCH
        tmp_inp = os.path.join(self.path, 'tmp.inp')
        with open(tmp_inp, 'w') as f:
            f.write('FINISH')

        # command must include "aas" flag to start MAPDL server
        command = '"%s" -aas -j %s -b -i tmp.inp -o out.txt -np %d %s' % (self._exec_file, self._jobname, self._nproc, self._additional_switches)

        # remove the broadcast file if it exists as the key will be
        # output here when ansys server is available
        if os.path.isfile(self._broadcast_file):
            os.remove(self._broadcast_file)

        # add run location to command
        self._log.debug('Spawning shell process with: "%s"', command)
        self._log.debug('At "%s"', self.path)

        # after v19, this is the only way this will work...
        if os.name == 'nt':
            command = 'START /B "MAPDL" %s' % command

        # set stdout
        if self._log.level < 20:  # < INFO            
            self._process = subprocess.Popen(command, shell=True, 
                                             cwd=self.path)
        else:
            self._process = subprocess.Popen(command, shell=True, 
                                             cwd=self.path,
                                             stdout=subprocess.PIPE,
                                             stderr=subprocess.PIPE)

        # listen for broadcast file
        self._log.debug('Waiting for valid key in %s', self._broadcast_file)
        telapsed = 0
        tstart = time.time()
        while telapsed < self._start_timeout:
            try:
                if os.path.isfile(self._broadcast_file):
                    with open(self._broadcast_file, 'r') as f:
                        text = f.read()
                        if 'visited:collaborativecosolverunitior' in text:
                            self._log.debug('Initialized ANSYS')
                            break
                time.sleep(0.1)
                telapsed = time.time() - tstart

            except KeyboardInterrupt:
                raise KeyboardInterrupt

        # exit if timed out
        if telapsed > self._start_timeout:
            err_msg = 'Unable to start ANSYS within %.1f seconds' % self._start_timeout
            self._log.error(err_msg)
            raise TimeoutError(err_msg)

        # open server
        keyfile = os.path.join(self.path, 'aaS_MapdlId.txt')
        with open(keyfile) as f:
            key = f.read()

        orb = CORBA.ORB_init()
        self._server = orb.string_to_object(key)

        # set to non-interactive
        # if os.name != 'nt':
            # text = self._server.executeCommandToString('/BATCH')

        try:
            self._server.getComponentName()
        except:
            raise RuntimeError('Unable to connect to APDL server')

        self._log.debug('Connected to ANSYS using CORBA interface with key %s', key)

        # separate logger for broadcast file
        if self._log_broadcast:
            self._broadcast_logger = self._start_broadcast_logger()

    @threaded
    def _start_broadcast_logger(self, update_rate=1.0):
        """ separate logger using broadcast_file """
        # listen to broadcast file
        loadstep = 0
        overall_progress = 0
        try:
            old_tail = ''
            old_size = 0
            while self.is_alive:
                new_size = os.path.getsize(self._broadcast_file)
                if new_size != old_size:
                    old_size = new_size
                    new_tail = tail(self._broadcast_file, 4)
                    if new_tail != old_tail:
                        lines = new_tail.split('>>')
                        for line in lines:
                            line = line.strip().replace('<<broadcast::', '')
                            if "current-load-step" in line:
                                n = int(re.search(r'\d+', line).group())
                                if n > loadstep:
                                    loadstep = n
                                    overall_progress = 0
                                    self._log.info(line)
                            elif "overall-progress" in line:
                                n = int(re.search(r'\d+', line).group())
                                if n > overall_progress:
                                    overall_progress = n
                                    self._log.info(line)
                        old_tail = new_tail
                time.sleep(update_rate)
        except Exception as e:
            pass

    def exit(self, close_log=True, timeout=3):
        """Exit MAPDL process"""
        self._log.debug('Exiting ANSYS')
        if self._server is not None:
            self._server.terminate()
            self._server = None

        if close_log:
            self._close_apdl_log()

        # wait for lockfile to be removed
        if timeout:
            tstart = time.time()
            while os.path.isfile(self._lockfile):
                time.sleep(0.05)
                telap = tstart - time.time()
                if telap > timeout:
                    return 1

    def kill(self):
        """Forces ANSYS process to end and removes lock file"""
        try:
            self.exit()
        except:
            pass

        if self._process is not None and os.name == 'linux':
            kill_process(self._process.pid)

        try:
            self._close_apdl_log()
        except:
            pass

        try:
            self._remove_lockfile()
        except:
            pass

    def _run(self, command):
        """Sends a command to the mapdl server via the CORBA interface"""
        self._reset_cache()
        if self._server is None:
            raise RuntimeError('ANSYS exited')

        # cleanup command
        command = command.strip()
        if not command:
            raise Exception('Empty command')

        if command[:4].lower() == 'cdre':
            with self.non_interactive:
                return self.run(command)

        if command[:4].lower() == '/com':
            split_command = command.split(',')
            if len(split_command) < 2:
                return ''
            elif not split_command[1]:
                return ''
            elif split_command[1]:
                if not split_command[1].strip():
                    return ''

        # /OUTPUT not redirected properly in corba
        if command[:4].lower() == '/out':
            items = command.split(',')
            if len(items) < 2:  # empty comment
                return ''
            elif not items[1]:  # empty comment
                return ''
            elif items[1]:
                if not items[1].strip():    # empty comment
                    return ''

            items = command.split(',')
            if len(items) > 1:  # redirect to file
                if len(items) > 2:
                    if items[2].strip():
                        filename = '.'.join(items[1:3]).strip()
                    else:
                        filename = '.'.join(items[1:2]).strip()
                else:
                    filename = items[1]

                if filename:
                    if os.path.basename(filename) == filename:
                        filename = os.path.join(self.path, filename)
                    self._output = filename
                    if len(items) == 5:
                        if items[4].lower().strip() == 'append':
                            self._outfile = open(filename, 'a')
                    else:
                        self._outfile = open(filename, 'w')
            else:
                self._output = ''
                if self._outfile:
                    self._outfile.close()
                self._outfile = None
            return ''

        # include error checking
        text = ''
        additional_text = ''

        self._log.debug('Running command %s' % command)
        text = self._server.executeCommandToString(command)

        # print supressed output
        additional_text = self._server.executeCommandToString('/GO')

        if 'is not a recognized' in text:
            if not self.allow_ignore:
                text = text.replace('This command will be ignored.', '')
                text += '\n\nIgnore these messages by setting allow_ignore=True'
                raise Exception(text)

        if text:
            text = text.replace('\\n', '\n')
            if '*** ERROR ***' in text:
                self._log.error(text)
                raise Exception(text)

        if additional_text:
            additional_text = additional_text.replace('\\n', '\n')
            if '*** ERROR ***' in additional_text:
                self._log.error(additional_text)
                raise Exception(additional_text)

        if self._interactive_plotting:
            self._display_plot('%s\n%s' % (text, additional_text))

        # return text, additional_text
        if text == additional_text:
            additional_text = ''
        return '%s\n%s' % (text, additional_text)



"""Module to read ANSYS ASCII block formatted CDB files"""
import os
import sys
import logging
from functools import wraps

import numpy as np
from vtk import (VTK_TETRA, VTK_QUADRATIC_TETRA, VTK_PYRAMID,
                 VTK_QUADRATIC_PYRAMID, VTK_WEDGE, VTK_QUADRATIC_WEDGE,
                 VTK_HEXAHEDRON, VTK_QUADRATIC_HEXAHEDRON, VTK_TRIANGLE, VTK_QUAD,
                 VTK_QUADRATIC_TRIANGLE, VTK_QUADRATIC_QUAD)
import vtk
import pyvista as pv

from pyansys import _reader
from pyansys.misc import vtk_cell_info
from pyansys.geometry import Geometry
from pyansys.cell_quality import quality

VTK9 = vtk.vtkVersion().GetVTKMajorVersion() >= 9

log = logging.getLogger(__name__)
log.setLevel('CRITICAL')



class Archive(Geometry):
    """Read a blocked ANSYS archive file.

    Reads a blocked CDB file and optionally parses it to a vtk grid.

    Write the archive file using ``CDWRITE, DB, archive.cdb``

    Parameters
    ----------
    filename : string
        Filename of block formatted cdb file

    read_parameters : bool, optional
        Optionally read parameters from the archive file.  Default
        ``False``.

    parse_vtk : bool, optional
        When ``True``, parse the raw data into to VTK format.

    force_linear : bool, optional
        This parser creates quadratic elements if available.  Set
        this to True to always create linear elements.  Defaults
        to False.

    allowable_types : list, optional
        Allowable element types.  Defaults to all valid element
        types in ``from pyansys.elements.valid_types``

        See help(pyansys.elements) for available element types.

    null_unallowed : bool, optional
        Elements types not matching element types will be stored
        as empty (null) elements.  Useful for debug or tracking
        element numbers.  Default False.

    Examples
    --------
    >>> import pyansys
    >>> hex_beam = pyansys.Archive(pyansys.examples.hexarchivefile)
    >>> print(hex_beam)
    ANSYS Archive File HexBeam.cdb
      Number of Nodes:              40
      Number of Elements:           321
      Number of Element Types:      1
      Number of Node Components:    2
      Number of Element Components: 2

    Print the node array

    >>> hex_beam.nodes
    array([[0.  , 0.  , 0.  ],
           [1.  , 0.  , 0.  ],
           [0.25, 0.  , 0.  ],
           ...,
           [0.75, 0.5 , 3.5 ],
           [0.75, 0.5 , 4.  ],
           [0.75, 0.5 , 4.5 ]])

    """

    def __init__(self, filename, read_parameters=False,
                 parse_vtk=True, force_linear=False,
                 allowable_types=None, null_unallowed=False, verbose=False):
        """Initializes an instance of the archive class."""
        self._read_parameters = read_parameters
        self._filename = filename
        self._raw = _reader.read(filename, read_parameters=read_parameters,
                                 debug=verbose)
        super().__init__(self._raw['nnum'],
                         self._raw['nodes'],
                         self._raw['elem'],
                         self._raw['elem_off'],
                         self._raw['ekey'],
                         node_comps=self._raw['node_comps'],
                         elem_comps=self._raw['elem_comps'],
                         rdat=self._raw['rdat'],
                         rnum=self._raw['rnum'],
                         keyopt=self._raw['keyopt'])

        if parse_vtk:
            self._grid = self._parse_vtk(allowable_types, force_linear, null_unallowed)

    @property
    def raw(self):
        raise AttributeError('The `raw` attribute has been depreciated.  Access'
                             ' the values directly from the archive object.\n\n'
                             '    Instead of:\n'
                             '    archive.raw["nodes"]\n'
                             '    \n'
                             '    Use\n'
                             '    print(archive.nodes)')

    @property
    def parameters(self):
        """Parameters stored in the archive file

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile,
                                      read_parameters=True)
        >>> archive.parameters
        {}
        """
        if not self._read_parameters:
            raise AttributeError('No parameters read.  Read the archive again '
                                 ' with ``read_parameters=True``')
        return self._raw['parameters']

    def __repr__(self):
        basename = os.path.basename(self._filename)
        txt = 'ANSYS Archive File %s\n' % basename
        txt += '  Number of Nodes:              %d\n' % len(self.nnum)
        txt += '  Number of Elements:           %d\n' % len(self.enum)
        txt += '  Number of Element Types:      %d\n' % len(self.ekey)
        txt += '  Number of Node Components:    %d\n' % len(self.node_components)
        txt += '  Number of Element Components: %d\n' % len(self.element_components)
        return txt

    @property
    def grid(self):
        """VTK and pyvista grid of the archive file.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.grid
        UnstructuredGrid (0x7ffa237f08a0)
          N Cells:      40
          N Points:     321
          X Bounds:     0.000e+00, 1.000e+00
          Y Bounds:     0.000e+00, 1.000e+00
          Z Bounds:     0.000e+00, 5.000e+00
          N Arrays:     13
        """
        if self._grid is None:
            raise AttributeError('Archive must be parsed as a vtk grid.\n'
                                 'Set `parse_vtk=True`')
        return self._grid

    @property
    def quality(self):
        """Minimum scaled jacobian cell quality.

        Negative values indicate invalid cells while positive values
        indicate valid cells.  Varies between -1 and 1.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.quality
        array([1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.,
               1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.,
               1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.])
        """
        if self._grid is None:
            raise AttributeError('Archive must be parsed as a vtk grid.\n'
                                 'Set `parse_vtk=True`')
        return quality(self._grid)

    @wraps(pv.plot)
    def plot(self, *args, **kwargs):
        """Plot the ANSYS archive file"""
        self.grid.plot(*args, **kwargs)


def chunks(l, n):
    """ Yield successive n-sized chunks from l """
    for i in range(0, len(l), n):
        yield l[i:i + n]


def save_as_archive(filename, grid, mtype_start=1, etype_start=1,
                    real_constant_start=1, mode='w', nblock=True,
                    enum_start=1, nnum_start=1,
                    include_etype_header=True, reset_etype=False,
                    allow_missing=True, include_surface_elements=True,
                    include_solid_elements=True):
    """Writes FEM as an ANSYS APDL archive file.  This function
    supports the following element types:

        - ``vtk.VTK_TETRA``
        - ``vtk.VTK_QUADRATIC_TETRA``
        - ``vtk.VTK_PYRAMID``
        - ``vtk.VTK_QUADRATIC_PYRAMID``
        - ``vtk.VTK_WEDGE``
        - ``vtk.VTK_QUADRATIC_WEDGE``
        - ``vtk.VTK_HEXAHEDRON``
        - ``vtk.VTK_QUADRATIC_HEXAHEDRON``

    Will automatically renumber nodes and elements if the FEM does not
    contain ANSYS node or element numbers.  Node numbers are stored as
    a point array "ansys_node_num", and cell numbers are stored as
    cell array "ansys_elem_num".

    Parameters
    ----------
    filename : str
       Filename to write archive file.

    grid : vtk.UnstructuredGrid
        VTK UnstructuredGrid to convert to an APDL archive file.
        PolyData will automatically be converted to an unstructured
        mesh.

    mtype_start : int, optional
        Material number to assign to elements.  Can be set manually by
        adding the cell array "mtype" to the unstructured grid.

    etype_start : int, optional
        Starting element type number.  Can be manually set by adding
        the cell array "ansys_etype" to the unstructured grid.

    real_constant_start : int, optional
        Starting real constant to assign to unset cells.  Can be
        manually set by adding the cell array "ansys_real_constant" to
        the unstructured grid.

    mode : str, optional
        File mode.  See help(open)

    nblock : bool, optional
        Write node block when writing archive file.

    enum_start : int, optional
        Starting element number to assign to unset cells.  Can be
        manually set by adding the cell array "ansys_elem_num" to the
        unstructured grid.

    nnum_start : int, optional
        Starting element number to assign to unset points.  Can be
        manually set by adding the point array "ansys_node_num" to the
        unstructured grid.

    include_etype_header : bool, optional
        For each element type, includes element type command
        (e.g. "ET, 1, 186") in the archive file.

    reset_etype : bool, optional
        Resets element type.  Element types will automatically be
        determined by the shape of the element (i.e. quadradic
        tetrahedrals will be saved as SOLID187, linear hexahedrals as
        SOLID185).  Default True.

    include_surface_elements : bool, optional
        Includes surface elements when writing the archive file and
        saves them as SHELL181.

    include_solid_elements : bool, optional
        Includes solid elements when writing the archive file and
        saves them as SOLID185, SOLID186, or SOLID187.

    """
    header = '/PREP7\n'

    if isinstance(grid, pv.PolyData):
        grid = grid.cast_to_unstructured_grid()

    allowable = []
    if include_solid_elements:
        allowable.extend([VTK_TETRA,
                          VTK_QUADRATIC_TETRA,
                          VTK_PYRAMID,
                          VTK_QUADRATIC_PYRAMID,
                          VTK_WEDGE,
                          VTK_QUADRATIC_WEDGE,
                          VTK_HEXAHEDRON,
                          VTK_QUADRATIC_HEXAHEDRON])

    if include_surface_elements:
        allowable.extend([VTK_TRIANGLE,
                          VTK_QUAD])
                          # VTK_QUADRATIC_TRIANGLE,
                          # VTK_QUADRATIC_QUAD

    # extract allowable cell types
    mask = np.in1d(grid.celltypes, allowable)
    grid = grid.extract_cells(mask)

    # node numbers
    if 'ansys_node_num' in grid.point_arrays:
        nodenum = grid.point_arrays['ansys_node_num']
    else:
        log.info('No ANSYS node numbers set in input.  ' +
                 'Adding default range')
        nodenum = np.arange(1, grid.number_of_points + 1)

    if np.any(nodenum == -1):
        if not allow_missing:
            raise Exception('Missing node numbers.  Exiting due "allow_missing=False"')
        start_num = nodenum.max() + 1
        if nnum_start > start_num:
            start_num = nnum_start
        nadd = np.sum(nodenum == -1)
        end_num = start_num + nadd
        log.info('FEM missing some node numbers.  Adding node numbering ' +
                 'from %d to %d' % (start_num, end_num))
        nodenum[nodenum == -1] = np.arange(start_num, end_num)

    # element block
    ncells = grid.number_of_cells
    if 'ansys_elem_num' in grid.cell_arrays:
        enum = grid.cell_arrays['ansys_elem_num']
    else:
        if not allow_missing:
            raise Exception('Missing node numbers.  Exiting due "allow_missing=False"')
        log.info('No ANSYS element numbers set in input.  ' +
                 'Adding default range starting from %d' % enum_start)
        enum = np.arange(1, ncells + 1)

    if np.any(enum == -1):
        if not allow_missing:
            raise Exception('-1 encountered in "ansys_elem_num".\n'
                            + 'Exiting due "allow_missing=False"')

        start_num = enum.max() + 1
        if enum_start > start_num:
           start_num = enum_start
        nadd = np.sum(enum == -1)
        end_num = start_num + nadd
        log.info('FEM missing some cell numbers.  Adding numbering ' +
                 'from %d to %d' % (start_num, end_num))
        enum[enum == -1] = np.arange(start_num, end_num)

    # material type
    if 'ansys_material_type' in grid.cell_arrays:
        mtype = grid.cell_arrays['ansys_material_type']
    else:
        log.info('No ANSYS element numbers set in input.  ' +
                 'Adding default range starting from %d' % mtype_start)
        mtype = np.arange(1, ncells + 1)

    if np.any(mtype == -1):
        log.info('FEM missing some material type numbers.  Adding...')
        mtype[mtype == -1] = mtype_start

    # real constant
    if 'ansys_real_constant' in grid.cell_arrays:
        rcon = grid.cell_arrays['ansys_real_constant']
    else:
        log.info('No ANSYS element numbers set in input.  ' +
                 'Adding default range starting from %d' % real_constant_start)
        rcon = np.arange(1, ncells + 1)

    if np.any(rcon == -1):
        log.info('FEM missing some material type numbers.  Adding...')
        rcon[rcon == -1] = real_constant_start

    # element type
    invalid = False
    if 'ansys_etype' in grid.cell_arrays and not reset_etype:
        missing = False
        typenum = grid.cell_arrays['ansys_elem_type_num']
        etype = grid.cell_arrays['ansys_etype']
        if np.any(etype == -1):
            log.warning('Some elements are missing element type numbers.')
            invalid = True

        if include_etype_header and not invalid:
            _, ind = np.unique(etype, return_index=True)
            for idx in ind:
                header += 'ET, %d, %d\n' % (etype[idx], typenum[idx])
    else:
        missing = True

    # check if valid
    if not missing:
        mask = grid.celltypes < 20
        if np.any(grid.cell_arrays['ansys_elem_type_num'][mask] == 186):
            invalid = True
            log.warning('Invalid ANSYS element types.')

    if invalid or missing:
        if not allow_missing:
            raise Exception('Invalid or missing data in "ansys_elem_type_num"' +
                            ' or "ansys_etype".  Exiting due "allow_missing=False"')

        log.info('No ANSYS element type or invalid data input.  ' +
                 'Adding default range starting from %d' % etype_start)

        etype = np.empty(grid.number_of_cells, np.int32)
        etype_185 = etype_start + 2
        etype[grid.celltypes == VTK_TETRA] = etype_185
        etype[grid.celltypes == VTK_HEXAHEDRON] = etype_185
        etype[grid.celltypes == VTK_WEDGE] = etype_185
        etype[grid.celltypes == VTK_PYRAMID] = etype_185

        etype_186 = etype_start
        etype[grid.celltypes == VTK_QUADRATIC_HEXAHEDRON] = etype_186
        etype[grid.celltypes == VTK_QUADRATIC_WEDGE] = etype_186
        etype[grid.celltypes == VTK_QUADRATIC_PYRAMID] = etype_186

        etype_187 = etype_start + 1
        etype[grid.celltypes == VTK_QUADRATIC_TETRA] = etype_187

        # Surface elements
        etype_181 = etype_start + 3
        etype[grid.celltypes == VTK_TRIANGLE] = etype_181
        etype[grid.celltypes == VTK_QUAD] = etype_181

        typenum = np.empty_like(etype)
        typenum[etype == etype_185] = 185
        typenum[etype == etype_186] = 186
        typenum[etype == etype_187] = 187
        typenum[etype == etype_181] = 181

        header += 'ET, %d, 185\n' % etype_185
        header += 'ET, %d, 186\n' % etype_186
        header += 'ET, %d, 187\n' % etype_187
        header += 'ET, %d, 181\n' % etype_181

    # number of nodes written per element
    elem_nnodes = np.empty(etype.size, np.int32)
    elem_nnodes[typenum == 181] = 4
    elem_nnodes[typenum == 185] = 8
    elem_nnodes[typenum == 186] = 20
    elem_nnodes[typenum == 187] = 10

    with open(str(filename), mode) as f:
        f.write(header)

        # write node block
        if write_nblock:
            write_nblock(f, nodenum, grid.points)

        # eblock header
        h = ''
        h += 'EBLOCK,19,SOLID,{:10d},{:10d}\n'.format(enum[-1], ncells)
        h += '(19i8)\n'
        f.write(h)

        celltypes = grid.celltypes
        cells, offset = vtk_cell_info(grid)
        if VTK9:
            offset += 1

        for i in range(ncells):
            if VTK9:
                nodes = nodenum[cells[offset[i]:offset[i + 1]]]
            else:
                c = offset[i]
                nnode = cells[c]
                c += 1
                nodes = nodenum[cells[c:c + nnode]]

            cellinfo = (mtype[i],          # Field 1: material reference number
                        etype[i],          # Field 2: element type number
                        rcon[i],           # Field 3: real constant reference number
                        1,                 # Field 4: section number
                        0,                 # Field 5: element coordinate system
                        0,                 # Field 6: Birth/death flag
                        0,                 # Field 7:
                        0,                 # Field 8:
                        elem_nnodes[i],    # Field 9: Number of nodes
                        0,                 # Field 10: Not Used
                        enum[i])           # Field 11: Element number
            line = '%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d' % cellinfo

            if celltypes[i] == VTK_QUADRATIC_TETRA:
                if typenum[i] == 187:
                    line += '%8d%8d%8d%8d%8d%8d%8d%8d\n%8d%8d\n' % tuple(nodes)
                else:  # must be 186
                    writenodes = (nodes[0],  # 0,  I
                                  nodes[1],  # 1,  J
                                  nodes[2],  # 2,  K
                                  nodes[2],  # 3,  L (duplicate of K)
                                  nodes[3],  # 4,  M
                                  nodes[3],  # 5,  N (duplicate of M)
                                  nodes[3],  # 6,  O (duplicate of M)
                                  nodes[3],  # 7,  P (duplicate of M)
                                  nodes[4],  # 8,  Q
                                  nodes[5],  # 9,  R
                                  nodes[3],  # 10, S (duplicate of K)
                                  nodes[6],  # 11, T
                                  nodes[3],  # 12, U (duplicate of M)
                                  nodes[3],  # 13, V (duplicate of M)
                                  nodes[3],  # 14, W (duplicate of M)
                                  nodes[3],  # 15, X (duplicate of M)
                                  nodes[7],  # 16, Y
                                  nodes[8],  # 17, Z
                                  nodes[9],  # 18, A
                                  nodes[9])  # 19, B (duplicate of A)

                    line += '%8d%8d%8d%8d%8d%8d%8d%8d\n' % writenodes[:8]
                    line += '%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d\n' % writenodes[8:]


            elif celltypes[i] == VTK_TETRA:
                writenodes = (nodes[0],  # 0,  I
                              nodes[1],  # 1,  J
                              nodes[2],  # 2,  K
                              nodes[2],  # 3,  L (duplicate of K)
                              nodes[3],  # 4,  M
                              nodes[3],  # 5,  N (duplicate of M)
                              nodes[3],  # 6,  O (duplicate of M)
                              nodes[3])  # 7,  P (duplicate of M)
                line += '%8d%8d%8d%8d%8d%8d%8d%8d\n' % writenodes

            elif celltypes[i] == VTK_WEDGE:
                writenodes = (nodes[2],  # 0,  I
                              nodes[1],  # 1,  J
                              nodes[0],  # 2,  K
                              nodes[0],  # 3,  L (duplicate of K)
                              nodes[5],  # 4,  M
                              nodes[4],  # 5,  N
                              nodes[3],  # 6,  O
                              nodes[3])  # 7,  P (duplicate of O)
                line += '%8d%8d%8d%8d%8d%8d%8d%8d\n' % writenodes

            elif celltypes[i] == VTK_QUADRATIC_WEDGE:
                writenodes = (nodes[2],  # 0,  I
                              nodes[1],  # 1,  J
                              nodes[0],  # 2,  K
                              nodes[0],  # 3,  L (duplicate of K)
                              nodes[5],  # 4,  M
                              nodes[4],  # 5,  N
                              nodes[3],  # 6,  O
                              nodes[3],  # 7,  P (duplicate of O)
                              nodes[7],  # 8,  Q
                              nodes[6],  # 9,  R
                              nodes[0],  # 10, S   (duplicate of K)
                              nodes[8],  # 11, T
                              nodes[10], # 12, U
                              nodes[9],  # 13, V
                              nodes[3],  # 14, W (duplicate of O)
                              nodes[11], # 15, X
                              nodes[14], # 16, Y
                              nodes[13], # 17, Z
                              nodes[12], # 18, A
                              nodes[12]) # 19, B (duplicate of A)
                line += '%8d%8d%8d%8d%8d%8d%8d%8d\n' % writenodes[:8]
                line += '%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d\n' % writenodes[8:]

            elif celltypes[i] == VTK_QUADRATIC_PYRAMID:
                writenodes = (nodes[0],  # 0,  I
                              nodes[1],  # 1,  J
                              nodes[2],  # 2,  K
                              nodes[3],  # 3,  L
                              nodes[4],  # 4,  M
                              nodes[4],  # 5,  N (duplicate of M)
                              nodes[4],  # 6,  O (duplicate of M)
                              nodes[4],  # 7,  P (duplicate of M)
                              nodes[5],  # 8,  Q
                              nodes[6],  # 9,  R
                              nodes[7],  # 10, S
                              nodes[8],  # 11, T
                              nodes[4],  # 12, U (duplicate of M)
                              nodes[4],  # 13, V (duplicate of M)
                              nodes[4],  # 14, W (duplicate of M)
                              nodes[4],  # 15, X (duplicate of M)
                              nodes[9],  # 16, Y
                              nodes[10], # 17, Z
                              nodes[11], # 18, A
                              nodes[12]) # 19, B (duplicate of A)

                line += '%8d%8d%8d%8d%8d%8d%8d%8d\n' % writenodes[:8]
                line += '%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d\n' % writenodes[8:]

            elif celltypes[i] == VTK_PYRAMID:
                writenodes = (nodes[0],  # 0,  I
                              nodes[1],  # 1,  J
                              nodes[2],  # 2,  K
                              nodes[3],  # 3,  L
                              nodes[4],  # 4,  M
                              nodes[4],  # 5,  N (duplicate of M)
                              nodes[4],  # 6,  O (duplicate of M)
                              nodes[4])  # 7,  P (duplicate of M)
                line += '%8d%8d%8d%8d%8d%8d%8d%8d\n' % writenodes[:8]

            elif celltypes[i] == VTK_HEXAHEDRON:
                line += '%8d%8d%8d%8d%8d%8d%8d%8d\n' % tuple(nodes[:8])

            elif celltypes[i] == VTK_QUADRATIC_HEXAHEDRON:
                line += '%8d%8d%8d%8d%8d%8d%8d%8d\n' % tuple(nodes[:8])
                line += '%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d\n' % tuple(nodes[8:])

            elif celltypes[i] == VTK_TRIANGLE:
                writenodes = (nodes[0],  # 0,  I
                              nodes[1],  # 1,  J
                              nodes[2],  # 2,  K
                              nodes[2])  # 3,  L (duplicate of K)
                line += '%8d%8d%8d%8d\n' % writenodes

            elif celltypes[i] == VTK_QUAD:
                writenodes = (nodes[0],  # 0,  I
                              nodes[1],  # 1,  J
                              nodes[2],  # 2,  K
                              nodes[3])  # 3,  L
                line += '%8d%8d%8d%8d\n' % writenodes

            # else:
            #     raise RuntimeError('Invalid write cell type %d' % celltypes[i])

            f.write(line)

        f.write('      -1\n')


def write_nblock(filename, node_id, pos, angles=None):
    """Writes nodes and node angles to file.

    Parameters
    ----------
    filename : str or file handle
        Filename to write node block to.

    node_id : np.ndarray
        ANSYS node numbers.

    pos : np.ndarray
        Node coordinates.

    angles : np.ndarray, optional
        Writes the node angles for each node when included.
    """
    assert pos.ndim == 2 and pos.shape[1] == 3, 'Invalid position array'
    if angles is not None:
        assert angles.ndim == 2 and angles.shape[1] == 3, 'Invalid angle array'

    # Header Tell ANSYS to start reading the node block with 6 fields,
    # associated with a solid, the maximum node number and the number
    # of lines in the node block
    h = '/PREP7 \n'
    h += 'NBLOCK,6,SOLID,%10d,%10d\n' % (np.max(node_id), pos.shape[0])
    h += '(3i8,6e20.13)'

    # NBLOCK footer
    f = 'N,R5.3,LOC,       -1, \n'

    # Sort input data
    ind = np.argsort(node_id)
    node_id = node_id[ind]
    pos = pos[ind]

    if angles is None:
        np.savetxt(
            filename,
            np.hstack((node_id.reshape(-1, 1), pos)),
            '%8d       0       0' +
            '%20.12E' *
            3,
            header=h,
            footer=f,
            comments='',
            newline='\n')
    else:
        # Array of node positions and angles
        arr = np.empty((node_id.size, 7))
        arr[:, 0] = node_id
        arr[:, 1:4] = pos
        arr[:, 4:] = angles

        # stack node IDs and positions
        np.savetxt(
            filename,
            arr,
            '%8d       0       0' +
            '%20.12E' *
            6,
            header=h,
            footer=f,
            comments='',
            newline='\n')


def write_cmblock(filename, items, comp_name, comp_type, digit_width=10):
    """Writes a component block, CMBLOCK, to a file.

    Parameters
    ----------
    filename : str or file handle
        File to write CMBLOCK component to

    items : list or np.ndarray
        Element or node numbers to write.

    comp_name : str
        Name of the component

    comp_type : str
        Component type to write.  Should be either 'ELEMENT' or 'NODE'.

    digit_width : int, optional
        Default 10
    """
    if comp_type.upper() not in ['ELEMENT', 'NODE']:
        raise ValueError("`comp_type` must be either 'ELEMENT' or 'NODE'")

    items = np.unique(items)

    toprint = []
    toprint.append(items[0])
    for i, value in enumerate(np.diff(items)):
        if value == 1:
            continue
        else:
            if items[i - 1] + 1 == items[i]:
                toprint.append(-items[i])
                toprint.append(items[i + 1])
            else:
                toprint.append(items[i + 1])

    # catch if last item is part of a list
    if toprint[-1] != abs(items[-1]):
        toprint.append(-items[i + 1])

    nitems = len(toprint)
    lines = []
    lines.append('CMBLOCK,%s,%s,%8d  ! from pyansys' % (comp_name.upper(),
                                                        comp_type.upper(),
                                                        nitems))
    lines.append('(8i%d)' % digit_width)
    digit_formatter = '%' + '%d' % digit_width + 'd'

    for chunk in chunks(toprint, 8):
        lines.append(''.join([digit_formatter] * len(chunk)) % tuple(chunk))

    lines.append('')

    # write file
    if sys.version_info[0] == 3:
        string_types = str
    else:
        string_types = basestring

    text = '\n'.join(lines)

    # either write to file or file object
    if isinstance(filename, string_types):
        open(filename, 'w').write(text)
    else:
        filename.write(text)



"""
Available as of 20.2:
LINK1
PLANE2
BEAM3
BEAM4
SOLID5
COMBIN7
LINK8
INFIN9
LINK10
LINK11
CONTAC12
PLANE13
COMBIN14
FLUID15
PIPE16
PIPE17
PIPE18
SURF19
PIPE20
MASS21
SURF22
BEAM23
BEAM24
PLANE25
MATRIX27
SHELL28
FLUID29
FLUID30
LINK31
LINK32
LINK33
LINK34
PLANE35
SOURC36
COMBIN37
FLUID38
COMBIN39
COMBIN40
SHELL41
PLANE42
SHELL43
BEAM44
SOLID45
SOLID46
INFIN47
MATRIX50
SHELL51
CONTAC52
PLANE53
BEAM54
PLANE55
SHELL57
PIPE59
PIPE60
SHELL61
SOLID62
SHELL63
SOLID64
SOLID65
FLUID66
PLANE67
LINK68
SOLID69
SOLID70
MASS71
PLANE75
PLANE77
PLANE78
FLUID79
FLUID80
FLUID81
PLANE82
PLANE83
SOLID87
VISCO88
VISCO89
SOLID90
SHELL91
SOLID92
SHELL93
CIRCU94
SOLID95
SOLID96
SOLID97
SOLID98
SHELL99
USER100
USER101
USER102
USER103
USER104
USER105
VISCO106
VISCO107
VISCO108
TRANS109
INFIN110
INFIN111
ROT112
INTER115
FLUID116
EDGE117
HF118
HF119
HF120
PLANE121
SOLID122
SOLID123
CIRCU124
CIRCU125
TRANS126
FLUID129
FLUID130
SHELL131
SHELL132
TRANS135
FLUID136
FLUID137
FLUID138
FLUID139
ROM140
SHELL143
ROM144
SURF151
SURF152
SURF153
SURF154
SURF155
SURF156
SHELL157
SURF159
TARGE169
TARGE170
CONTA171
CONTA172
CONTA173
CONTA174
CONTA175
CONTA176
CONTA177
CONTA178
PRETS179
LINK180
SHELL181
PLANE182
PLANE183
RBAR184
SOLID185
SOLID186
SOLID187
BEAM188
BEAM189
SOLID190
SOLID191
INTER192
INTER193
INTER194
INTER195
LAYER196
LAYER197
FOLLW201
INTER202
INTER203
INTER204
INTER205
INTER206
SHELL208
SHELL209
CPT212
CPT213
COMBI214
CPT215
CPT216
CPT217
FLUID218
FLUID219
FLUID220
FLUID221
PLANE222
SOLID223
SOLID225
SOLID226
SOLID227
PLANE230
SOLID231
SOLID232
PLANE233
SOLID236
SOLID237
PLANE238
SOLID239
SOLID240
HSFLD241
HSFLD242
COMBI250
SURF251
SURF252
INFIN257
INFIN258
INFIN259
INFIN260
GLINK262
REINF263
REINF264
REINF265
SOLID272
SOLID273
SOLID278
SOLID279
CABLE280
SHELL281
SHELL282
SHELL283
SOLID285
PIPE288
PIPE289
ELBOW290
SOLID291
PLANE292
PLANE293
USER300
"""
import numpy as np

# element type to VTK conversion function call map
# 0: skip
# 1: Point
# 2: Line (linear or quadratic)
# 3: Shell
# 4: 3D Solid (Hexahedral, wedge, pyramid, tetrahedral)
# 5: Tetrahedral
# 6: Line (always linear)

_etype_map = [0, 2,  # LINK1
	      3,  # PLANE2
	      3,  # BEAM3
	      2,  # BEAM4
	      4,  # SOLID5
	      0,  # UNUSED6
	      1,  # COMBIN7
	      2,  # LINK8
	      0,  # INFIN9
	      2,  # LINK10
	      2,  # LINK11
	      2,  # CONTAC12
	      3,  # PLANE13
	      2,  # COMBIN14
	      0,  # FLUID15
	      2,  # PIPE16
	      2,  # PIPE17
	      2,  # PIPE18
	      0,  # SURF19
	      2,  # PIPE20
	      1,  # MASS21
	      3,  # SURF22
	      2,  # BEAM23
	      2,  # BEAM24
	      3,  # PLANE25
	      0,  # UNUSED26
	      0,  # MATRIX27
	      3,  # SHELL28
	      3,  # FLUID29
	      4,  # FLUID30
	      2,  # LINK31
	      2,  # LINK32
	      2,  # LINK33
	      2,  # LINK34
	      3,  # PLANE35
	      0,  # SOURC36
	      2,  # COMBIN37
	      2,  # FLUID38
	      2,  # COMBIN39
	      2,  # COMBIN40
	      3,  # SHELL41
	      3,  # PLANE42
	      3,  # SHELL43
	      2,  # BEAM44
	      4,  # SOLID45
	      4,  # SOLID46
	      0,  # INFIN47
	      0,  # UNUSED48
	      0,  # UNUSED49
	      0,  # MATRIX50
	      3,  # SHELL51
	      0,  # CONTAC52
	      3,  # PLANE53
	      3,  # BEAM54
	      3,  # PLANE55
	      0,  # UNUSED56
	      3,  # SHELL57
	      0,  # UNUSED58
	      2,  # PIPE59
	      2,  # PIPE60
	      2,  # SHELL61
	      4,  # SOLID62
	      3,  # SHELL63
	      4,  # SOLID64
	      4,  # SOLID65
	      2,  # FLUID66
	      3,  # PLANE67
	      2,  # LINK68
	      4,  # SOLID69
	      4,  # SOLID70
	      1,  # MASS71
	      0,  # UNUSED72
	      0,  # UNUSED73
	      0,  # UNUSED74
	      3,  # PLANE75
	      0,  # UNUSED76
	      3,  # PLANE77
	      3,  # PLANE78
	      3,  # FLUID79
	      4,  # FLUID80
	      3,  # FLUID81
	      3,  # PLANE82
	      3,  # PLANE83
	      0,  # UNUSED84
	      0,  # UNUSED85
	      0,  # UNUSED86
	      5,  # SOLID87
	      3,  # VISCO88
	      4,  # VISCO89
	      4,  # SOLID90
	      3,  # SHELL91
	      5,  # SOLID92
	      3,  # SHELL93
	      0,  # CIRCU94
	      4,  # SOLID95
	      4,  # SOLID96
	      4,  # SOLID97
	      5,  # SOLID98
	      3,  # SHELL99
	      4,  # USER100
	      4,  # USER101
	      4,  # USER102
	      4,  # USER103
	      4,  # USER104
	      4,  # USER105
	      3,  # VISCO106
	      4,  # VISCO107
	      4,  # VISCO108
	      0,  # TRANS109
	      0,  # INFIN110
	      0,  # INFIN111
	      0,  # ROT112
	      0,  # UNUSED113
	      0,  # UNUSED114
	      3,  # INTER115
	      3,  # FLUID116
	      4,  # EDGE117
	      3,  # HF118
	      5,  # HF119
	      4,  # HF120
	      3,  # PLANE121
	      4,  # SOLID122
	      5,  # SOLID123
	      0,  # CIRCU124
	      0,  # CIRCU125
	      2,  # TRANS126
	      0,  # UNUSED127
	      0,  # UNUSED128
	      2,  # FLUID129
	      3,  # FLUID130
	      3,  # SHELL131
	      3,  # SHELL132
	      0,  # UNUSED133
	      0,  # UNUSED134
	      0,  # TRANS135
	      3,  # FLUID136
	      0,  # FLUID137
	      0,  # FLUID138
	      0,  # FLUID139
	      5,  # ROM140
	      0,  # UNUSED141
	      0,  # UNUSED142
	      3,  # SHELL143
	      0,  # ROM144
	      0,  # UNUSED145
	      0,  # UNUSED146
	      0,  # UNUSED147
	      0,  # UNUSED148
	      0,  # UNUSED149
	      0,  # UNUSED150
	      2,  # SURF151
	      3,  # SURF152
	      2,  # SURF153
	      3,  # SURF154
	      3,  # SURF155
	      2,  # SURF156
	      3,  # SHELL157
	      0,  # UNUSED158
	      0,  # SURF159
	      0,  # UNUSED160
	      2,  # BEAM161
	      0,  # UNUSED162
	      3,  # SHELL163
	      4,  # SOLID164
	      0,  # UNUSED165
	      0,  # UNUSED166
	      0,  # UNUSED167
	      5,  # SOLID168
	      0,  # TARGE169
	      0,  # TARGE170
	      2,  # CONTA171
	      2,  # CONTA172
	      3,  # CONTA173
	      3,  # CONTA174
	      1,  # CONTA175
	      2,  # CONTA176
	      2,  # CONTA177
	      2,  # CONTA178
	      0,  # PRETS179
	      2,  # LINK180
	      3,  # SHELL181
	      3,  # PLANE182
	      3,  # PLANE183
	      0,  # RBAR184
	      4,  # SOLID185
	      4,  # SOLID186
	      5,  # SOLID187
	      6,  # BEAM188
	      2,  # BEAM189
	      4,  # SOLSH190
              0,  # UNUSED191
	      4,  # INTER192
	      0,  # INTER193
	      0,  # INTER194
	      0,  # INTER195
	      0,  # LAYER196
	      0,  # LAYER197
	      0,  # UNUSED198
	      0,  # UNUSED199
	      0,  # UNUSED200
	      0,  # FOLLW201
	      0,  # INTER202
	      0,  # INTER203
	      0,  # INTER204
	      0,  # INTER205
	      0,  # INTER206
	      0,  # UNUSED207
	      2,  # SHELL208
	      2,  # SHELL209
	      0,  # UNUSED210
	      0,  # UNUSED211
	      0,  # CPT212
	      0,  # CPT213
	      0,  # COMBI214
	      0,  # CPT215
              0,  # UNUSED216
	      5,  # CPT217
	      3,  # FLUID218
	      3,  # FLUID219
	      4,  # FLUID220
	      5,  # FLUID221
	      3,  # PLANE222
	      3,  # PLANE223
	      0,  # UNUSED224
	      0,  # SOLID225
	      4,  # SOLID226
	      5,  # SOLID227
	      0,  # UNUSED228
	      0,  # UNUSED229
	      3,  # PLANE230
	      4,  # SOLID231
	      5,  # SOLID232
	      3,  # PLANE233
	      0,  # UNUSED234
	      0,  # UNUSED235
	      4,  # SOLID236
	      5,  # SOLID237
	      3,  # PLANE238
	      4,  # SOLID239
	      5,  # SOLID240
	      0,  # HSFLD241
	      0,  # HSFLD242
	      0,  # UNUSED243
	      0,  # UNUSED244
	      0,  # UNUSED245
	      0,  # UNUSED246
	      0,  # UNUSED247
	      0,  # UNUSED248
	      0,  # UNUSED249
	      2,  # COMBI250
	      2,  # SURF251
	      3,  # SURF252
	      0,  # UNUSED253
	      0,  # UNUSED254
	      0,  # UNUSED255
	      0,  # UNUSED256
	      0,  # INFIN257
	      0,  # INFIN258
	      0,  # INFIN259
              0,  # UNUSED260
	      4,  # UNUSED261
	      0,  # GLINK262
	      0,  # REINF263
	      0,  # REINF264
	      0,  # REINF265
	      0,  # UNUSED266
	      0,  # UNUSED267
	      0,  # UNUSED268
	      0,  # UNUSED269
	      0,  # UNUSED270
	      0,  # UNUSED271
	      0,  # SOLID272
	      0,  # SOLID273
	      0,  # UNUSED274
	      0,  # UNUSED275
	      0,  # UNUSED276
	      0,  # UNUSED277
	      4,  # SOLID278
	      4,  # SOLID279
	      2,  # CABLE280
	      3,  # SHELL281
	      3,  # SHELL282
	      3,  # SHELL283
	      0,  # UNUSED284
	      5,  # SOLID285
	      0,  # UNUSED286
	      0,  # UNUSED287
	      2,  # PIPE288
	      2,  # PIPE289
	      2,  # ELBOW290
	      5,  # SOLID291
	      3,  # PLANE292
	      3,  # PLANE293
	      0,  # UNUSED294
	      0,  # UNUSED295
	      0,  # UNUSED296
	      0,  # UNUSED297
	      0,  # UNUSED298
	      0,  # UNUSED299
	      0,  # USER300
]

ETYPE_MAP = np.array(_etype_map, dtype=np.int32)




"""
Used fortran header file for item definitions
/usr/ansys_inc/v150/ansys/customize/include/fdfull.inc

"""
import os
import warnings

import numpy as np

from pyansys import _binary_reader
from pyansys.common import (read_table, AnsysBinary, parse_header, two_ints_to_long,
                            read_standard_header)


FRONTAL_FULL_HEADER_KEYS = [
    'fun04', 'neqn', 'nmrow', 'nmatrx', 'kan',
    'wfmax', 'lenbac', 'numdof', 'jcgtrmL', 'jcgtrmH',
    'lumpm', 'jcgeqn', 'jcgtrm', 'keyuns', 'extopt',
    'keyse', 'sclstf', 'nxrows', 'ptrIDXl', 'ptrIDXh',
    'ncefull', 'ncetrm', 'ptrENDl', 'ptrENDh', '0'
]

SYMBOLIC_FULL_HEADER_KEYS = [
    'fun04', 'neqn', 'nmrow', 'nmatrx', 'kan',
    'wfmax', 'lenbac', 'numdof', 'ntermKl', 'ntermKh',  # (10)
    'lumpm', 'nmrow', 'ntermK_', 'keyuns', 'extopt',
    'keyse', 'sclstf', 'nxrows', 'ptrSTFl', 'ptrSTFh',  # (20)
    'ncefull', 'ntermMh', 'ptrENDl', 'ptrENDh', 'ptrIRHSl',
    'ptrIRHSh', 'ptrMASl', 'ptrMASh', 'ptrDMPl', 'ptrDMPh',  # (30)
    'ptrCEl', 'ptrCEh', 'nNodes', 'ntermMl', 'ntermDl',
    'ptrDOFl', 'ptrDOFh', 'ptrRHSl', 'ptrRHSh', 'ntermDh',   # (40)
    'ngMaxNZ', 'ptrNGPHl', 'ptrNGPHh', 'minKdiag', 'maxKdiag',
    'minMdiag', 'maxMdiag', 'minDdiag', 'maxDdiag', 'ngTerml',  # (50)
    'ngTermh', 'ngTermCl', 'ngTermCh', 'ptrDIAGKl', 'ptrDIAGKh',
    'ptrDIAGMl', 'ptrDIAGMh', 'ptrDIAGCl', 'ptrDIAGCh', 'ptrSCLKl',  # (60)
    'ptrSCLKh', 'Glbneqn', 'distKey', 'ngTermFl', 'ngTermFh',
    'GlbnNodes', 'GlbnVars', 'GlbfAcCE', 'lcAcLen', 'GlbfCE',   # (70)
    'ptrGmtl', 'ptrGmth', 'nceGprime', 'numA12A11', 'strctChg',
    'ntermGl', 'ntermGh', 'ptrDensel', 'ptrDenseh', 'nVirtBCs',  # (80)
    'ptrVrtBCl', 'ptrVrtBCh', 'ptrMRKl', 'ptrMRKh'
]


class FullFile(AnsysBinary):
    """Stores the results of an ANSYS full file.

    Parameters
    ----------
    filename : str
        Filename of the full file to read.

    Examples
    --------
    >>> import pyansys
    >>> full = pyansys.read_binary('file.rst')
    >>> print(full)
    PyANSYS - MAPDL Full File
    Title                    : Demo
    Version                  : 20.1
    Platform                 : WINDOWS x64
    Jobname                  : file
    Matrices                 : 2
    Equations                : 345
    Nodes                    : 115
    Degrees of Freedom       : 3

    """

    def __init__(self, filename):
        """Loads full header on initialization.

        See ANSYS programmer's reference manual full header section
        for definitions of each header.

        """
        self._const = None
        self._krow = None
        self._kcol = None
        self._kdata = None
        self._mrow = None
        self._mcol = None
        self._mdata = None
        self._k = None
        self._m = None
        self._dof_ref = None

        self.filename = filename
        self._standard_header = read_standard_header(self.filename)
        self._header = parse_header(self.read_record(103), SYMBOLIC_FULL_HEADER_KEYS)

        # if not self._header['fun04'] < 0:
            # raise NotImplementedError("Unable to read a frontal assembly full file")

        # Check if lumped (item 11)
        if self._header['lumpm']:
            raise NotImplementedError("Unable to read a lumped mass matrix")

        # Check if arrays are unsymmetric (item 14)
        if self._header['keyuns']:
            raise NotImplementedError("Unable to read an unsymmetric mass/stiffness matrix")

    @property
    def k(self):
        """Stiffness Matrix corresponding to sorted DOF.

        Examples
        --------
        >>> import pyansys
        >>> full = pyansys.read_binary('file.rst')
        >>> print(full.k)
        <345x345 sparse matrix of type '<class 'numpy.float64'>'
                with 7002 stored elements in Compressed Sparse Column format>
        """
        if self._k is None:
            self._load_km()
        return self._k

    @property
    def m(self):
        """Mass Matrix corresponding to sorted DOF.

        Examples
        --------
        >>> import pyansys
        >>> full = pyansys.read_binary('file.rst')
        >>> full.m
        <345x345 sparse matrix of type '<class 'numpy.float64'>'
                with 2883 stored elements in Compressed Sparse Column format>
        """
        if self._m is None:
            self._load_km()
        return self._m

    @property
    def dof_ref(self):
        """Sorted degree of freedom reference.

        Examples
        --------
        >>> import pyansys
        >>> full = pyansys.read_binary('file.rst')
        >>> full.dof_ref
        array([[  1,   0],
               [  1,   1],
               [  1,   2],
               [115,   0],
               [115,   1],
               [115,   2]], dtype=int32)

        Notes
        -----
        Obtain the unsorted degree of freedom reference with
        >>> dof_ref, k, m = sparse_full.load_km(sort=False)
        """
        if self._dof_ref is None:
            self._load_km()
        return self._dof_ref

    @property
    def const(self):
        """Constrained DOF
        Returns the node number and DOF constrained in ANSYS.

        Examples
        --------
        >>> full.const
        array([], shape=(0, 2), dtype=int32)
        """
        if self._const is None:
            self._load_km()
        return self._const

    def load_km(self, as_sparse=True, sort=False):
        """Load and construct mass and stiffness matrices from an
        ANSYS full file.

        Parameters
        ----------
        as_sparse : bool, optional
            Outputs the mass and stiffness matrices as scipy csc
            sparse arrays when True by default.

        sort : bool, optional
            Rearranges the k and m matrices such that the rows
            correspond to to the sorted rows and columns in dor_ref.
            Also sorts dor_ref.

        Returns
        -------
        dof_ref : (n x 2) np.int32 array
            This array contains the node and degree corresponding to
            each row and column in the mass and stiffness matrices.
            In a 3 DOF analysis the dof integers will correspond to:
            0 - x
            1 - y
            2 - z
            Sort these values by node number and DOF by enabling the
            sort parameter.

        k : (n x n) np.float or scipy.csc array
            Stiffness array

        m : (n x n) np.float or scipy.csc array
            Mass array

        Examples
        --------
        >>> import pyansys
        >>> full = pyansys.read_binary('file.rst')
        >>> dof_ref, k, m = full.load_km()
        >>> print(k)
        (0, 0)       163408119.6581276
        (0, 1)               0.0423270
        (1, 1)       163408119.6581276
        :	:
        (342, 344)     6590544.8717949
        (343, 344)    -6590544.8717950
        (344, 344)    20426014.9572689

        Notes
        -----
        Constrained entries are removed from the mass and stiffness
        matrices.

        Constrained DOF can be accessed from ``const``, which returns
        the node number and DOF constrained in ANSYS.
        """
        if not os.path.isfile(self.filename):
            raise Exception('%s not found' % self.filename)

        if as_sparse:
            try:
                from scipy.sparse import csc_matrix, coo_matrix
            except ImportError:
                raise ImportError('Unable to load scipy, use ``load_km`` with '
                                  '``as_sparse=False``')

        # Get header details
        neqn = self._header['neqn']  # Number of equations

        # number of terms in stiffness matrix
        ntermK = two_ints_to_long(self._header['ntermKl'], self._header['ntermKh'])

        ptrSTF = self._header['ptrSTF']  # Location of stiffness matrix
        ptrMAS = self._header['ptrMAS']  # Location in file to mass matrix

        # number of terms in mass matrix
        ntermM = two_ints_to_long(self._header['ntermMl'], self._header['ntermMh'])
        ptrDOF = self._header['ptrDOF']  # pointer to DOF info

        # DOF information
        with open(self.filename, 'rb') as f:
            read_table(f, skip=True)  # standard header
            read_table(f, skip=True)  # full header
            read_table(f, skip=True)  # number of degrees of freedom

            # Nodal equivalence table
            neqv = read_table(f, cython=True)

            # read number of degrees of freedom for each node and constant tables
            f.seek(ptrDOF*4)
            ndof = read_table(f, cython=True)
            const = read_table(f, cython=True)

        # degree of freedom reference and number of degress of freedom per node
        dof_ref = [ndof, neqv]
        self.ndof = ndof

        # Read k and m blocks (see help(ReadArray) for block description)
        if ntermK:
            krow, kcol, kdata = _binary_reader.read_array(self.filename,
                                                          ptrSTF,
                                                          ntermK,
                                                          neqn,
                                                          const)
        else:
            warnings.warn('Missing stiffness matrix')
            kdata = None

        if ntermM:
            mrow, mcol, mdata = _binary_reader.read_array(self.filename,
                                                          ptrMAS,
                                                          ntermM,
                                                          neqn,
                                                          const)
        else:
            warnings.warn('Missing mass matrix')
            mdata = None

        # remove constrained entries
        if np.any(const < 0):
            if kdata is not None:
                remove = np.nonzero(const < 0)[0]
                mask = ~np.logical_or(np.in1d(krow, remove), np.in1d(kcol, remove))
                krow = krow[mask]
                kcol = kcol[mask]
                kdata = kdata[mask]

            if mdata is not None:
                mask = ~np.logical_or(np.in1d(mrow, remove), np.in1d(mcol, remove))
                mrow = mrow[mask]
                mcol = mcol[mask]
                mdata = mdata[mask]


        # sort nodal equivalence
        dof_ref, index, nref, dref = _binary_reader.sort_nodal_eqlv(neqn, neqv, ndof)

        # store constrained dof information
        unsort_dof_ref = np.vstack((nref, dref)).T
        self._const = unsort_dof_ref[const < 0]

        if sort:  # make sorting the same as ANSYS rdfull would output
            # resort to make in upper triangle
            krow = index[krow]
            kcol = index[kcol]
            krow, kcol = np.sort(np.vstack((krow, kcol)), 0)

            if mdata is not None:
                mrow = index[mrow]
                mcol = index[mcol]
                mrow, mcol = np.sort(np.vstack((mrow, mcol)), 0)

        else:
            dof_ref = unsort_dof_ref

        # store data for later reference
        if kdata is not None:
            self._krow = krow
            self._kcol = kcol
            self._kdata = kdata
        if mdata is not None:
            self._mrow = mrow
            self._mcol = mcol
            self._mdata = mdata

        # output as a sparse matrix
        if as_sparse:

            if kdata is not None:
                k = coo_matrix((neqn,) * 2)
                k.data = kdata  # data has to be set first
                k.row = krow
                k.col = kcol

                # convert to csc matrix (generally faster for sparse solvers)
                k = csc_matrix(k)
            else:
                k = None

            if mdata is not None:
                m = coo_matrix((neqn,) * 2)
                m.data = mdata
                m.row = mrow
                m.col = mcol

                # convert to csc matrix (generally faster for sparse solvers)
                m = csc_matrix(m)
            else:
                m = None

        else:
            if kdata is not None:
                k = np.zeros((neqn,) * 2)
                k[krow, kcol] = kdata
            else:
                k = None

            if mdata is not None:
                m = np.zeros((neqn,) * 2)
                m[mrow, mcol] = mdata
            else:
                m = None

        return dof_ref, k, m

    def _load_km(self):
        """Loads the matrices with sorted DOF"""
        self._dof_ref, self._k, self._m = self.load_km(sort=True)

    def __str__(self):
        rst_info = ['PyANSYS - MAPDL Full File']

        def add_info(key, value):
            rst_info.append('{:<25s}: {:s}'.format(key, str(value)))

        keys = ['title', 'subtitle']
        for key in keys:
            value = self._standard_header[key]
            if value:
                add_info(key.capitalize(), value)

        add_info('Version', self._standard_header['verstring'])
        add_info('Platform', self._standard_header['machine'])
        add_info('Jobname', self._standard_header['jobname'])
        add_info('Matrices', self._header['nmatrx'])
        add_info('Equations', self._header['neqn'])
        add_info('Nodes', self._header['nNodes'])
        add_info('Degrees of Freedom', self._header['numdof'])

        return '\n'.join(rst_info)
